{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/convert/CleanVBIN.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2023-11-07 20:14:42 +0100 (Di, 07 Nov 2023) $
  Revision: $Rev: 675 $

  Visionaire Engine cleanup mechanism for binary XMLs
}
unit CleanVBIN;

interface

uses
  Shared;

type
  TBinaryXMLNode = packed record
    NameID     : Word;
    NumChildren: Word;
  end;

  PXMLNode = ^TXMLNode;
  TXMLNode = record
    BinaryXMLNode: TBinaryXMLNode;
    Children     : array of PXMLNode;
  end;

function CleanBinaryXML(const BINBuffer: TBuffer;
  out CleanBuffer: TBuffer): Boolean;

implementation

uses
  {$IFNDEF VIS_VERSION_5}
  Names3,
  {$ELSE}
  Names5,
  {$ENDIF}
  VEDMap,
  SysUtils,
  StrUtils;

var
  VBINHasVersions: Boolean = False;

function ReadContent(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean; forward;

function ReadBasic_Int(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: Longint): Boolean;
begin
  Result := False;

  if BINIdx + SizeOf(Val) > Length(BINBuffer) then
    Exit;

  Move(BINBuffer[BINIdx], Val, SizeOf(Val));
  Inc(BINIdx, SizeOf(Val));

  Result := True
end;

function ReadBasic_Float(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: Single): Boolean;
begin
  Result := False;

  if BINIdx + SizeOf(Val) > Length(BINBuffer) then
    Exit;

  Move(BINBuffer[BINIdx], Val, SizeOf(Val));
  Inc(BINIdx, SizeOf(Val));

  Result := True
end;

function ReadBasic_Bool(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: Boolean): Boolean;
var
  ValByte: Byte;
begin
  Result := False;

  if BINIdx + SizeOf(ValByte) > Length(BINBuffer) then
    Exit;

  Move(BINBuffer[BINIdx], ValByte, SizeOf(ValByte));
  Inc(BINIdx, SizeOf(ValByte));

  Val := Boolean(ValByte);

  Result := True
end;

function ReadBasic_String(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: UTF8String): Boolean;
var
  Len: Longint;
begin
  // get string length first
  Result := ReadBasic_Int(BINBuffer, BINIdx, Len);
  if not Result then
    Exit;

  // get the actual string content
  if Len > 0 then
  begin
    SetLength(Val, Len);

    Result := (BINIdx + Len <= Length(BINBuffer));
    if not Result then
      Exit;

    Move(BINBuffer[BINIdx], Val[1], Len);
    Inc(BINIdx, Len);

    // the string may be terminated by a trailing zero character, get rid of it
    SetLength(Val, Length(StrPas(@Val[1])))
  end
  else
    Val := '';

  Result := True
end;

procedure WriteBasic_Int(var CleanBuffer: TBuffer; var CleanIdx: Longint;
  const Val: Longint);
begin
  Move(Val, CleanBuffer[CleanIdx], SizeOf(Val));
  Inc(CleanIdx, SizeOf(Val))
end;

procedure WriteBasic_Float(var CleanBuffer: TBuffer; var CleanIdx: Longint;
  const Val: Single);
begin
  Move(Val, CleanBuffer[CleanIdx], SizeOf(Val));
  Inc(CleanIdx, SizeOf(Val))
end;

procedure WriteBasic_Bool(var CleanBuffer: TBuffer; var CleanIdx: Longint;
  const Val: Boolean);
var
  ValByte: Byte;
begin
  ValByte := Byte(Val);

  Move(ValByte, CleanBuffer[CleanIdx], SizeOf(ValByte));
  Inc(CleanIdx, SizeOf(ValByte))
end;

procedure WriteBasic_String(var CleanBuffer: TBuffer; var CleanIdx: Longint;
  Val: UTF8String);
var
  Len: Longint;
begin
  // prepare string: append zero terminator
  if Length(Val) > 0 then
    Val := Val + #000;

  // write string length first
  Len := Length(Val);
  WriteBasic_Int(CleanBuffer, CleanIdx, Len);

  // write the actual string content
  if Len > 0 then
  begin
    Move(Val[1], CleanBuffer[CleanIdx], Len);
    Inc(CleanIdx, Len)
  end
end;

function Read_Int(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  Val: Longint;
begin
  Result := ReadBasic_Int(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, Val)
end;

function Read_Float(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  Val: Single;
begin
  Result := ReadBasic_Float(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  WriteBasic_Float(CleanBuffer, CleanIdx, Val)
end;

function Read_Bool(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  Val: Boolean;
begin
  Result := ReadBasic_Bool(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  WriteBasic_Bool(CleanBuffer, CleanIdx, Val)
end;

function Read_String(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  Val   : UTF8String;
  ArcDes: AnsiString;

  procedure PerformCleanup(SearchPattern: AnsiString);
  var
    p: Longint;
  begin
    p := 1;
    repeat
      p := PosEx(SearchPattern, Val, p);

      if p <= 0 then
        Break;

      if p > ArchiveDesignatorLength then
        ArcDes := Copy(Val, p - ArchiveDesignatorLength + 1,
          ArchiveDesignatorLength)
      else
        ArcDes := '';

      if IsArchiveDesignator(ArcDes) then
      begin
        Val := Copy(Val, 1, p - ArchiveDesignatorLength) +
          Copy(Val, p + 1, Length(Val) - p);
        p := p - ArchiveDesignatorLength - 1
      end;

      Inc(p)
    until p > Length(Val) - 1;
  end;

begin
  Result := ReadBasic_String(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  // Do the first cleanup operation: If the string ends in a valid archive
  // designator, we cut it away.
  if Length(Val) > ArchiveDesignatorLength then
    ArcDes := Copy(Val, Length(Val) - ArchiveDesignatorLength + 1,
      ArchiveDesignatorLength)
  else
    ArcDes := '';

  if IsArchiveDesignator(ArcDes) then
    Val := Copy(Val, 1, Length(Val) - ArchiveDesignatorLength);

  // Do the second cleanup operation: Scripts contained in strings might contain
  // references to files, complete with an archive designator. These should be
  // enclosed in quotes.
  PerformCleanup('#"');
  PerformCleanup('#''');

  WriteBasic_String(CleanBuffer, CleanIdx, Val)
end;

function Read_Path(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  Val   : UTF8String;
  ArcDes: AnsiString;
begin
  Result := ReadBasic_String(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  // Do the cleanup operation here: If the path ends in a valid archive
  // designator, we cut it away.
  if Length(Val) > ArchiveDesignatorLength then
    ArcDes := Copy(Val, Length(Val) - ArchiveDesignatorLength + 1,
      ArchiveDesignatorLength)
  else
    ArcDes := '';

  if IsArchiveDesignator(ArcDes) then
    Val := Copy(Val, 1, Length(Val) - ArchiveDesignatorLength);

  WriteBasic_String(CleanBuffer, CleanIdx, Val)
end;

function Read_Point(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i: Longint;
begin
  // create node's children
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, 2);

  for i := 0 to Length(ParentXMLNode.Children) - 1 do
    New(ParentXMLNode.Children[i]);

  ParentXMLNode.Children[0].BinaryXMLNode.NameID := 1;
  ParentXMLNode.Children[0].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[1].BinaryXMLNode.NameID := 2;
  ParentXMLNode.Children[1].BinaryXMLNode.NumChildren := 0;

  // read children's contents
  for i := 0 to Length(ParentXMLNode.Children) - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end;

  Result := True
end;

function Read_Rect(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i: Longint;
begin
  // create node's children
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, 4);

  for i := 0 to Length(ParentXMLNode.Children) - 1 do
    New(ParentXMLNode.Children[i]);

  ParentXMLNode.Children[0].BinaryXMLNode.NameID := 4;
  ParentXMLNode.Children[0].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[1].BinaryXMLNode.NameID := 3;
  ParentXMLNode.Children[1].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[2].BinaryXMLNode.NameID := 6;
  ParentXMLNode.Children[2].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[3].BinaryXMLNode.NameID := 5;
  ParentXMLNode.Children[3].BinaryXMLNode.NumChildren := 0;

  // read children's contents
  for i := 0 to Length(ParentXMLNode.Children) - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end;

  Result := True
end;

function Read_Sprite(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i: Longint;
begin
  // create node's children
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, 6);

  for i := 0 to Length(ParentXMLNode.Children) - 1 do
    New(ParentXMLNode.Children[i]);

  ParentXMLNode.Children[0].BinaryXMLNode.NameID := 7;
  ParentXMLNode.Children[0].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[1].BinaryXMLNode.NameID := 32;
  ParentXMLNode.Children[1].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[2].BinaryXMLNode.NameID := 12;
  ParentXMLNode.Children[2].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[3].BinaryXMLNode.NameID := 8;
  ParentXMLNode.Children[3].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[4].BinaryXMLNode.NameID := 50;
  ParentXMLNode.Children[4].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[5].BinaryXMLNode.NameID := 51;
  ParentXMLNode.Children[5].BinaryXMLNode.NumChildren := 0;

  // read children's contents
  for i := 0 to Length(ParentXMLNode.Children) - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end;

  Result := True
end;

function Read_Text(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i: Longint;
begin
  // create node's children: due to the double usage of T209 (as table and
  // string attribute), we read the node contents explicitly rather than passing
  // the task to ReadContent
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, 3);

  for i := 0 to Length(ParentXMLNode.Children) - 1 do
    New(ParentXMLNode.Children[i]);

  ParentXMLNode.Children[0].BinaryXMLNode.NameID := 209;
  ParentXMLNode.Children[0].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[1].BinaryXMLNode.NameID := 7;
  ParentXMLNode.Children[1].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[2].BinaryXMLNode.NameID := 27;
  ParentXMLNode.Children[2].BinaryXMLNode.NumChildren := 0;

  Result := Read_String(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
    CleanIdx);
  if not Result then
    Exit;

  Result := Read_Path(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer, CleanIdx);
  if not Result then
    Exit;

  Result := Read_Int(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer, CleanIdx);
  if not Result then
    Exit;

  Result := True
end;

function Read_VInt(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 27;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VFloat(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 34;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VBool(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 44;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VString(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 25;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VPath(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 7;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VPoint(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 24;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VRect(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 22;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VSprite(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 29;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_VText(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i,
  VectorLen: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  // create a child node for each element and read its contents
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, VectorLen);

  for i := 0 to VectorLen - 1 do
  begin
    New(ParentXMLNode.Children[i]);
    ParentXMLNode.Children[i].BinaryXMLNode.NameID := 26;
    ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0;

    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end
end;

function Read_Link(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i: Longint;
begin
  // create node's children
  ParentXMLNode.BinaryXMLNode.NumChildren := 0;
  SetLength(ParentXMLNode.Children, 4);

  for i := 0 to Length(ParentXMLNode.Children) - 1 do
    New(ParentXMLNode.Children[i]);

  ParentXMLNode.Children[0].BinaryXMLNode.NameID := 13;
  ParentXMLNode.Children[0].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[1].BinaryXMLNode.NameID := 19;
  ParentXMLNode.Children[1].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[2].BinaryXMLNode.NameID := 36;
  ParentXMLNode.Children[2].BinaryXMLNode.NumChildren := 0;
  ParentXMLNode.Children[3].BinaryXMLNode.NameID := 21;
  ParentXMLNode.Children[3].BinaryXMLNode.NumChildren := 0;

  // read children's contents
  for i := 0 to Length(ParentXMLNode.Children) - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end;

  Result := True
end;

function Read_Links(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;

  procedure DuplicateNode(const SourceXMLNode: PXMLNode;
    var TargetXMLNode: PXMLNode);
  var
    i: Longint;
  begin
    New(TargetXMLNode);

    TargetXMLNode^.BinaryXMLNode := SourceXMLNode^.BinaryXMLNode;
    SetLength(TargetXMLNode^.Children, Length(SourceXMLNode^.Children));

    for i := 0 to Length(TargetXMLNode^.Children) - 1 do
      DuplicateNode(SourceXMLNode^.Children[i], TargetXMLNode^.Children[i])
  end;

var
  i,
  VectorLen,
  VersionedId: Longint;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

  if (VectorLen = 101010101) or (VectorLen = 101010102) then
  begin
    if VectorLen = 101010102 then
    begin
      // This links table has been versioned; get the versionedId attribute
      // value.
      Result := ReadBasic_Int(BINBuffer, BINIdx, VersionedId);
      if not Result then
        Exit;

      WriteBasic_Int(CleanBuffer, CleanIdx, VersionedId)
    end;

    // direct object table: read the number of rows in the table
    Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
    if not Result then
      Exit;

    WriteBasic_Int(CleanBuffer, CleanIdx, VectorLen);

    // an object table has to have exactly one child that serves as a structure
    // template
    Result := (ParentXMLNode.BinaryXMLNode.NumChildren = 1);
    if not Result then
      Exit;

    if VectorLen > 1 then
    begin
      // duplicate child (take note that one child already exists)
      SetLength(ParentXMLNode.Children, VectorLen);

      for i := 1 to VectorLen - 1 do
        DuplicateNode(ParentXMLNode.Children[0], ParentXMLNode.Children[i])
    end
    else if VectorLen = 0 then
      // remove existing child
      SetLength(ParentXMLNode.Children, 0)
  end
  else
  begin
    // regular links table: create a child node for each element
    ParentXMLNode.BinaryXMLNode.NumChildren := 0;
    SetLength(ParentXMLNode.Children, VectorLen);

    for i := 0 to VectorLen - 1 do
    begin
      New(ParentXMLNode.Children[i]);
      ParentXMLNode.Children[i].BinaryXMLNode.NameID := 15;
      ParentXMLNode.Children[i].BinaryXMLNode.NumChildren := 0
    end;

    Result := True
  end
end;

function Read_Table(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
begin
  Result := True
end;

function ReadNode(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i: Longint;
begin
  Result := False;

  SetLength(ParentXMLNode.Children, ParentXMLNode.BinaryXMLNode.NumChildren);

  // process all children
  for i := 0 to Length(ParentXMLNode.Children) - 1 do
  begin
    if BINIdx + SizeOf(ParentXMLNode.BinaryXMLNode) > Length(BINBuffer) then
      Exit;

    New(ParentXMLNode.Children[i]);

    // get the current child and copy it to the output
    Move(BINBuffer[BINIdx], ParentXMLNode.Children[i]^.BinaryXMLNode,
      SizeOf(ParentXMLNode.Children[i]^.BinaryXMLNode));
    Inc(BINIdx, SizeOf(ParentXMLNode.Children[i]^.BinaryXMLNode));
    Move(ParentXMLNode.Children[i]^.BinaryXMLNode, CleanBuffer[CleanIdx],
      SizeOf(ParentXMLNode.Children[i]^.BinaryXMLNode));
    Inc(CleanIdx, SizeOf(ParentXMLNode.Children[i]^.BinaryXMLNode));

    // do recursive descent for elements as well as attributes
    if not ReadNode(BINBuffer, BINIdx, ParentXMLNode.Children[i]^, CleanBuffer,
      CleanIdx) then
      Exit
  end;

  Result := True
end;

function ReadContent(const BINBuffer: TBuffer; var BINIdx: Longint;
  var ParentXMLNode: TXMLNode; var CleanBuffer: TBuffer;
  var CleanIdx: Longint): Boolean;
var
  i      : Longint;
  VisType: TVisType;
begin
  Result := True;

  // Workaround: If the content starts with a special value, we will have to
  // read an additional integer later on. The purpose of this construct is not
  // completely clear, but it might be related to the VBIN versioning scheme.
  if ParentXMLNode.BinaryXMLNode.NameID = 100 then
  begin
    Result := ReadBasic_Int(BINBuffer, BINIdx, i);
    if i = 101010102 then
    begin
      WriteBasic_Int(CleanBuffer, CleanIdx, i);
      VBINHasVersions := True
    end
    else
      Dec(BINIdx, 4)
  end;

  VisType := MapVisType(ParentXMLNode.BinaryXMLNode.NameID);

  // prepare link table
  if VisType = VT_Links then
  begin
    Result := Read_Links(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
      CleanIdx);
    if not Result then
      Exit
  end;

  // process current node's children first
  for i := 0 to Length(ParentXMLNode.Children) - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentXMLNode.Children[i]^,
      CleanBuffer, CleanIdx);
    if not Result then
      Exit
  end;

  // read node details depending on type
  case VisType of
    VT_Int:
      Result := Read_Int(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Float:
      Result := Read_Float(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Bool:
      Result := Read_Bool(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_String:
      Result := Read_String(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Path:
      Result := Read_Path(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Point:
      Result := Read_Point(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Rect:
      Result := Read_Rect(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Sprite:
      Result := Read_Sprite(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Text:
      Result := Read_Text(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VInt:
      Result := Read_VInt(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VFloat:
      Result := Read_VFloat(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VBool:
      Result := Read_VBool(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VString:
      Result := Read_VString(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VPath:
      Result := Read_VPath(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VPoint:
      Result := Read_VPoint(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VRect:
      Result := Read_VRect(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VSprite:
      Result := Read_VSprite(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_VText:
      Result := Read_VText(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Link:
      Result := Read_Link(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
    VT_Links:
      ; // already processed above
    VT_Table:
      Result := Read_Table(BINBuffer, BINIdx, ParentXMLNode, CleanBuffer,
        CleanIdx);
  else
    WriteLn(Format('    warning: type %d unsupported', [Longint(VisType)]));
  end;

  // If the content started with a special value (see above), we will have to
  // read (and discard) an additional integer after the version attribute. The
  // purpose of this value is not yet clear, but it might be related to the
  // VBIN versioning scheme.
  if (ParentXMLNode.BinaryXMLNode.NameID = 17) and VBINHasVersions then
  begin
    Result := ReadBasic_Int(BINBUffer, BINIdx, i);
    if not Result then
      Exit;

    WriteBasic_Int(CleanBuffer, CleanIdx, i)
  end
end;

procedure DestroyNode(var XMLNode: TXMLNode);
var
  i: Longint;
begin
  if Length(XMLNode.Children) > 0 then
  begin
    for i := 0 to High(XMLNode.Children) do
    begin
      DestroyNode(XMLNode.Children[i]^);
      Dispose(XMLNode.Children[i]);
      XMLNode.Children[i] := nil
    end;

    Finalize(XMLNode.Children);
    SetLength(XMLNode.Children, 0)
  end;
end;

function CleanBinaryXML(const BINBuffer: TBuffer;
  out CleanBuffer: TBuffer): Boolean;
var
  StructIdx,
  CleanIdx : Longint;
  XMLNode  : TXMLNode;
begin
  Result := False;

  // Prepare the output buffer: It will never get larger than the input.
  CleanIdx := 0;
  SetLength(CleanBuffer, Length(BINBuffer));
  FillChar(CleanBuffer[0], Length(CleanBuffer), 0);

  // prepare the root node
  StructIdx := 0;
  if StructIdx + SizeOf(XMLNode.BinaryXMLNode) > Length(BINBuffer) then
  begin
    WriteLn('    error parsing document structure');
    Exit
  end;

  // read node data and copy it to the output
  Move(BINBuffer[StructIdx], XMLNode.BinaryXMLNode,
    SizeOf(XMLNode.BinaryXMLNode));
  Inc(StructIdx, SizeOf(XMLNode.BinaryXMLNode));
  Move(XMLNode.BinaryXMLNode, CleanBuffer[CleanIdx],
    SizeOf(XMLNode.BinaryXMLNode));
  Inc(CleanIdx, SizeOf(XMLNode.BinaryXMLNode));

  // read the document tree structure recursively
  if not ReadNode(BINBuffer, StructIdx, XMLNode, CleanBuffer, CleanIdx) then
  begin
    WriteLn('    error parsing document structure');
    Exit
  end;

  // read the document content recursively
  if not ReadContent(BINBuffer, StructIdx, XMLNode, CleanBuffer, CleanIdx) then
  begin
    WriteLn('    error parsing document content');
    Exit
  end;

  // clean up temporary node structures
  DestroyNode(XMLNode);

  // shorten the output buffer
  SetLength(CleanBuffer, CleanIdx);

  Result := True
end;

end.
