{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/convert/ConvVBIN.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2023-11-07 20:14:42 +0100 (Di, 07 Nov 2023) $
  Revision: $Rev: 675 $

  Visionaire Engine converter for binary XMLs
}
unit ConvVBIN;

interface

uses
  Shared,
  NativeXml;

type
  TBinaryXMLNode = packed record
    NameID     : Word;
    NumChildren: Word;
  end;

function ConvertVBINToStructuredXML(const BINBuffer: TBuffer;
  out XMLDocument: TNativeXml; DoMapping: Boolean): Boolean;
function ConvertVBINToXML(var BINBuffer: TBuffer;
  out XMLBuffer: TBuffer; DoMapping: Boolean): Boolean;

implementation

uses
  VEDMap,
  {$IFNDEF CONVERSION_RESTRICTED}
  VEDMapEx,
  {$ENDIF}
  Classes,
  SysUtils;

const
  // preallocation factor for memory streams to keep the process from
  // reallocating (which consumes time and additional memory, possibly leading
  // to out of memory exceptions)
  StreamPreallocFactor  = 5;
  StreamPreallocDivisor = 2;

var
  VBINHasVersions: Boolean = False;

function GetTemporaryNodeName(NameID: Word): AnsiString;
begin
  Result := Format('%d', [NameID])
end;

function GetFinalNodeName(NameID: Word; DoMapping: Boolean): AnsiString;
begin
  {$IFNDEF CONVERSION_RESTRICTED}
  if DoMapping then
    Result := MapIdentifier(Format('T%d', [NameID]))
  else
    Result := Format('T%d', [NameID])
  {$ELSE}
  Result := Format('T%d', [NameID])
  {$ENDIF}
end;

function ReadContent(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean; forward;

function ReadBasic_Int(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: Longint): Boolean;
begin
  Result := False;

  if BINIdx + SizeOf(Val) > Length(BINBuffer) then
    Exit;

  Move(BINBuffer[BINIdx], Val, SizeOf(Val));
  Inc(BINIdx, SizeOf(Val));

  Result := True
end;

function ReadBasic_Float(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: Single): Boolean;
begin
  Result := False;

  if BINIdx + SizeOf(Val) > Length(BINBuffer) then
    Exit;

  Move(BINBuffer[BINIdx], Val, SizeOf(Val));
  Inc(BINIdx, SizeOf(Val));

  Result := True
end;

function ReadBasic_Bool(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: Boolean): Boolean;
var
  ValByte: Byte;
begin
  Result := False;

  if BINIdx + SizeOf(ValByte) > Length(BINBuffer) then
    Exit;

  Move(BINBuffer[BINIdx], ValByte, SizeOf(ValByte));
  Inc(BINIdx, SizeOf(ValByte));

  Val := Boolean(ValByte);

  Result := True
end;

function ReadBasic_String(const BINBuffer: TBuffer; var BINIdx: Longint;
  out Val: UTF8String): Boolean;
var
  Len: Longint;
begin
  // get string length first
  Result := ReadBasic_Int(BINBuffer, BINIdx, Len);
  if not Result then
    Exit;

  // get the actual string content
  if Len > 0 then
  begin
    SetLength(Val, Len);

    Result := (BINIdx + Len <= Length(BINBuffer));
    if not Result then
      Exit;

    Move(BINBuffer[BINIdx], Val[1], Len);
    Inc(BINIdx, Len);

    // the string may be terminated by a trailing zero character, get rid of it
    SetLength(Val, Length(StrPas(@Val[1])))
  end
  else
    Val := '';

  Result := True
end;

function Read_Int(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode): Boolean;
var
  Val: Longint;
begin
  Result := ReadBasic_Int(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  ParentNode.Value := Format('%d', [Val])
end;

function Read_Float(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode): Boolean;
var
  Val   : Single;
  DecSep: Char;
begin
  Result := ReadBasic_Float(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  DecSep := DecimalSeparator;
  DecimalSeparator := '.';
  ParentNode.Value := Format('%-1.6f', [Val]);
  DecimalSeparator := DecSep
end;

function Read_Bool(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode): Boolean;
var
  Val: Boolean;
begin
  Result := ReadBasic_Bool(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  if Val then
    ParentNode.Value := 'T'
  else
    ParentNode.Value := 'F'
end;

function Read_String(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode): Boolean;
var
  Val: UTF8String;
begin
  Result := ReadBasic_String(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  ParentNode.Value := Val
end;

function Read_Path(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode): Boolean;
var
  Val: UTF8String;
begin
  Result := ReadBasic_String(BINBuffer, BINIdx, Val);
  if not Result then
    Exit;

  ParentNode.Value := Val
end;

function Read_Point(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i   : Longint;
  Node: TXmlNode;
begin
  // create node's children
  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(1);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(2);

  // read children's contents
  for i := 0 to ParentNode.NodeCount - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentNode.Nodes[i], DoMapping);
    if not Result then
      Exit
  end;

  Result := True
end;

function Read_Rect(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i   : Longint;
  Node: TXmlNode;
begin
  // create node's children

  // The first two children actually contain the rect's "x" and "y" position. We
  // convert these to "left" and "top".
  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(4);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(3);

  // The next two attributes actually contain the rect's width and height. We
  // convert these to "right" and "bottom" afterwards
  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(6);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(5);

  // read children's contents
  for i := 0 to ParentNode.NodeCount - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentNode.Nodes[i], DoMapping);
    if not Result then
      Exit
  end;

  // do conversion as explained above: "right" ...
  ParentNode.Nodes[2].Value := Format('%d',
    [ParentNode.Nodes[2].ValueAsInteger + ParentNode.Nodes[0].ValueAsInteger]);

  // ... and "bottom"
  ParentNode.Nodes[3].Value := Format('%d',
    [ParentNode.Nodes[3].ValueAsInteger + ParentNode.Nodes[1].ValueAsInteger]);

  Result := True
end;

function Read_Sprite(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i   : Longint;
  Node: TXmlNode;
begin
  // create node's children
  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(7);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(32);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(12);

  Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(8);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(50);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(51);

  // read children's contents
  for i := 0 to ParentNode.NodeCount - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentNode.Nodes[i], DoMapping);
    if not Result then
      Exit
  end;

  Result := True
end;

function Read_Text(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  Node: TXmlNode;
begin
  // create node's children: due to the double usage of T209 (as table and
  // string attribute), we read the node contents explicitly rather than passing
  // the task to ReadContent
  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetFinalNodeName(209, DoMapping);
  Result := Read_String(BINBuffer, BINIdx, Node);
  if not Result then
    Exit;

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetFinalNodeName(7, DoMapping);
  Result := Read_Path(BINBuffer, BINIdx, Node);
  if not Result then
    Exit;

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetFinalNodeName(27, DoMapping);
  Result := Read_Int(BINBuffer, BINIdx, Node);
  if not Result then
    Exit;

  Result := True
end;

function Read_VInt(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node,
  Attr     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetFinalNodeName(27, DoMapping);

    // each (basic) vector element contains an attribute with the same name as
    // the node itself
    Attr := TsdAttribute.CreateParent(Node.Document, Node);
    Attr.Name := GetTemporaryNodeName(27);

    Result := ReadContent(BINBuffer, BINIdx, Attr, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VFloat(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node,
  Attr     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetFinalNodeName(34, DoMapping);

    // each (basic) vector element contains an attribute with the same name as
    // the node itself
    Attr := TsdAttribute.CreateParent(Node.Document, Node);
    Attr.Name := GetTemporaryNodeName(34);

    Result := ReadContent(BINBuffer, BINIdx, Attr, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VBool(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node,
  Attr     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetFinalNodeName(44, DoMapping);

    // each (basic) vector element contains an attribute with the same name as
    // the node itself
    Attr := TsdAttribute.CreateParent(Node.Document, Node);
    Attr.Name := GetTemporaryNodeName(44);

    Result := ReadContent(BINBuffer, BINIdx, Attr, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VString(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node,
  Attr     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetFinalNodeName(25, DoMapping);

    // each (basic) vector element contains an attribute with the same name as
    // the node itself
    Attr := TsdAttribute.CreateParent(Node.Document, Node);
    Attr.Name := GetTemporaryNodeName(25);

    Result := ReadContent(BINBuffer, BINIdx, Attr, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VPath(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node,
  Attr     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetFinalNodeName(7, DoMapping);

    // each (basic) vector element contains an attribute with the same name as
    // the node itself
    Attr := TsdAttribute.CreateParent(Node.Document, Node);
    Attr.Name := GetTemporaryNodeName(7);

    Result := ReadContent(BINBuffer, BINIdx, Attr, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VPoint(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetTemporaryNodeName(24);

    Result := ReadContent(BINBuffer, BINIdx, Node, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VRect(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetTemporaryNodeName(22);

    Result := ReadContent(BINBuffer, BINIdx, Node, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VSprite(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetTemporaryNodeName(29);

    Result := ReadContent(BINBuffer, BINIdx, Node, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_VText(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen: Longint;
  Node     : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  // create a child node for each element and read its contents
  for i := 1 to VectorLen do
  begin
    Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetTemporaryNodeName(26);

    Result := ReadContent(BINBuffer, BINIdx, Node, DoMapping);
    if not Result then
      Exit
  end
end;

function Read_Link(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i   : Longint;
  Node: TXmlNode;
begin
  // create node's children
  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(13);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(19);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(36);

  Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
  Node.Name := GetTemporaryNodeName(21);

  // read children's contents
  for i := 0 to ParentNode.NodeCount - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentNode.Nodes[i], DoMapping);
    if not Result then
      Exit
  end;

  Result := True
end;

function Read_Links(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i,
  VectorLen,
  NameID,
  VersionedId: Longint;
  IsVersioned: Boolean;
  Node       : TXmlNode;
begin
  // get number of elements
  Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
  if not Result then
    Exit;

  if (VectorLen = 101010101) or (VectorLen = 101010102) then
  begin
    IsVersioned := (VectorLen = 101010102);

    if IsVersioned then
    begin
      // This links table has been versioned; get the versionedId attribute
      // value.
      Result := ReadBasic_Int(BINBuffer, BINIdx, VersionedId);
      if not Result then
        Exit
    end;

    // direct object table: read the number of rows in the table
    Result := ReadBasic_Int(BINBuffer, BINIdx, VectorLen);
    if not Result then
      Exit;

    // an object table has to have exactly one child that serves as a structure
    // template
    Result := (ParentNode.NodeCount = 1);
    if not Result then
      Exit;

    if VectorLen > 1 then
      // duplicate child (take note that one child already exists)
      for i := 1 to VectorLen - 1 do
        (TsdElement.CreateParent(ParentNode.Document, ParentNode)).Assign(
          ParentNode.Nodes[0])
    else if VectorLen = 0 then
      // remove existing child
      ParentNode.NodesClear;

    // determine the current table's name
    if not TryStrToInt(ParentNode.Name, NameID) then
      Exit;

    // add attribute "versionedId"
    if IsVersioned then
    begin
      Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
      Node.Name := GetFinalNodeName(52, DoMapping);
      Node.ValueAsInteger := VersionedId
    end;

    // add attribute "tableId"
    Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode);
    Node.Name := GetFinalNodeName(19, DoMapping);
    Node.ValueAsInteger := MapTableId(NameID)
  end
  else
    // regular links table: create a child node for each element and read their
    // contents (Note: This should normally be done in ReadContent, but due to
    // the fact that the ParentNode there is a const-Parameter, the routine does
    // not notice the increase in NodeCount ... This is actually not nice at
    // all, but I won't fix it for now.)
    for i := 1 to VectorLen do
    begin
      Node := TsdElement.CreateParent(ParentNode.Document, ParentNode);
      Node.Name := GetTemporaryNodeName(15);

      Result := ReadContent(BINBuffer, BINIdx, Node, DoMapping);
      if not Result then
        Exit
    end
end;

function Read_Table(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode): Boolean;
begin
  Result := True
end;

function ReadNode(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentXMLNode: TBinaryXMLNode; const ParentNode: TXmlNode): Boolean;
var
  i            : Longint;
  BinaryXMLNode: TBinaryXMLNode;
  VisType      : TVisType;
  Node         : TXmlNode;
begin
  Result := False;

  // process all children
  for i := 0 to ParentXMLNode.NumChildren - 1 do
  begin
    if BINIdx + SizeOf(BinaryXMLNode) > Length(BINBuffer) then
      Exit;

    // get the current child
    Move(BINBuffer[BINIdx], BinaryXMLNode, SizeOf(BinaryXMLNode));
    Inc(BINIdx, SizeOf(BinaryXMLNode));

    // create the current child
    VisType := MapVisType(BinaryXMLNode.NameID);

    if VisType in BasicVisTypes then
      Node := TsdAttribute.CreateParent(ParentNode.Document, ParentNode)
    else
    begin
      if not (VisType in StructuredVisTypes) then
        WriteLn(Format('    warning: type %d unsupported', [Longint(VisType)]));

      Node := TsdElement.CreateParent(ParentNode.Document, ParentNode)
    end;

    // temporarily name the node
    Node.Name := GetTemporaryNodeName(BinaryXMLNode.NameID);

    // do recursive descent for elements as well as attributes
    if not ReadNode(BINBuffer, BINIdx, BinaryXMLNode, Node) then
      Exit
  end;

  Result := True
end;

function ReadContent(const BINBuffer: TBuffer; var BINIdx: Longint;
  const ParentNode: TXmlNode; DoMapping: Boolean): Boolean;
var
  i      : Longint;
  NameID : Longint;
  VisType: TVisType;
begin
  Result := True;

  // determine the current node's type
  if not TryStrToInt(ParentNode.Name, NameID) then
    Exit;

  // Workaround: If the content starts with a special value, we will have to
  // read an additional integer later on. The purpose of this construct is not
  // completely clear, but it might be related to the VBIN versioning scheme.
  if NameID = 100 then
  begin
    Result := ReadBasic_Int(BINBuffer, BINIdx, i);
    if i = 101010102 then
      VBINHasVersions := True
    else
      Dec(BINIdx, 4)
  end;

  VisType := MapVisType(NameID);

  // prepare link table
  if VisType = VT_Links then
  begin
    Result := Read_Links(BINBuffer, BINIdx, ParentNode, DoMapping);
    if not Result then
      Exit
  end;

  // set the final node name
  ParentNode.Name := GetFinalNodeName(NameID, DoMapping);

  // process current node's children first
  for i := 0 to ParentNode.NodeCount - 1 do
  begin
    Result := ReadContent(BINBuffer, BINIdx, ParentNode.Nodes[i], DoMapping);
    if not Result then
      Exit
  end;

  // read node details depending on type
  case VisType of
    VT_Int:
      Result := Read_Int(BINBuffer, BINIdx, ParentNode);
    VT_Float:
      Result := Read_Float(BINBuffer, BINIdx, ParentNode);
    VT_Bool:
      Result := Read_Bool(BINBuffer, BINIdx, ParentNode);
    VT_String:
      Result := Read_String(BINBuffer, BINIdx, ParentNode);
    VT_Path:
      Result := Read_Path(BINBuffer, BINIdx, ParentNode);
    VT_Point:
      Result := Read_Point(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_Rect:
      Result := Read_Rect(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_Sprite:
      Result := Read_Sprite(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_Text:
      Result := Read_Text(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VInt:
      Result := Read_VInt(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VFloat:
      Result := Read_VFloat(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VBool:
      Result := Read_VBool(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VString:
      Result := Read_VString(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VPath:
      Result := Read_VPath(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VPoint:
      Result := Read_VPoint(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VRect:
      Result := Read_VRect(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VSprite:
      Result := Read_VSprite(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_VText:
      Result := Read_VText(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_Link:
      Result := Read_Link(BINBuffer, BINIdx, ParentNode, DoMapping);
    VT_Links:
      ; // already processed above
    VT_Table:
      Result := Read_Table(BINBuffer, BINIdx, ParentNode);
  else
    WriteLn(Format('    warning: type %d unsupported', [Longint(VisType)]));
  end;

  // If the content started with a special value (see above), we will have to
  // read (and discard) an additional integer after the version attribute. The
  // purpose of this value is not yet clear, but it might be related to the
  // VBIN versioning scheme.
  if (NameID = 17) and VBINHasVersions then
  begin
    Result := ReadBasic_Int(BINBUffer, BINIdx, i);
    if not Result then
      Exit
  end
end;

procedure FixAttributes(const ParentNode: TXmlNode);
var
  i,
  NodeIdx : Longint;
  NodeList: array of TXmlNode;
begin
  NodeIdx := 0;
  SetLength(NodeList, ParentNode.NodeCount);

  // collect all pseudo-attributes and recursively process element nodes
  for i := 0 to ParentNode.NodeCount - 1 do
  begin
    if ParentNode.Nodes[i] is TsdAttribute then
    begin
      NodeList[NodeIdx] := ParentNode.Nodes[i];
      Inc(NodeIdx)
    end
    else
      FixAttributes(ParentNode.Nodes[i])
  end;

  SetLength(NodeList, NodeIdx);

  // move pseudo-attributes to their rightful place ...
  for i := 0 to NodeIdx - 1 do
  begin
    ParentNode.NodeExtract(NodeList[i]);
    ParentNode.AttributeAdd(TsdAttribute(NodeList[i]))
  end;

  SetLength(NodeList, 0)
end;

function ConvertVBINToStructuredXML(const BINBuffer: TBuffer;
  out XMLDocument: TNativeXml; DoMapping: Boolean): Boolean;
var
  StructIdx    : Longint;
  BinaryXMLNode: TBinaryXMLNode;
begin
  Result := False;

  // start a new XML document
  XMLDocument := TNativeXML.Create(nil);
  XMLDocument.XmlFormat := xfReadable;

  // prepare the root node
  StructIdx := 0;
  if StructIdx + SizeOf(BinaryXMLNode) > Length(BINBuffer) then
  begin
    WriteLn('    error parsing document structure');
    Exit
  end;

  Move(BINBuffer[StructIdx], BinaryXMLNode, SizeOf(BinaryXMLNode));
  Inc(StructIdx, SizeOf(BinaryXMLNode));

  // temporarily name the root
  XMLDocument.Root.Name := GetTemporaryNodeName(BinaryXMLNode.NameID);

  // read the document tree structure recursively
  if not ReadNode(BINBuffer, StructIdx, BinaryXMLNode, XMLDocument.Root) then
  begin
    WriteLn('    error parsing document structure');
    Exit
  end;

  // read the document content recursively
  if not ReadContent(BINBuffer, StructIdx, XMLDocument.Root, DoMapping) then
  begin
    WriteLn('    error parsing document content');
    Exit
  end;

  // Up to now all attributes are somewhat in limbo due to the way they were
  // created -- which we needed, since this preserved the node order. Convert
  // them to real attributes in the sense of TNativeXml now.
  FixAttributes(XMLDocument.Root);

  Result := True
end;

function ConvertVBINToXML(var BINBuffer: TBuffer;
  out XMLBuffer: TBuffer; DoMapping: Boolean): Boolean;
var
  XMLDocument : TNativeXml;
  XMLStream   : TMemoryStream;
  PreallocSize: Longint;
begin
  Result := ConvertVBINToStructuredXML(BINBuffer, XMLDocument, DoMapping);
  if not Result then
    Exit;

  PreallocSize := (StreamPreallocFactor * Length(BINBuffer))
    div StreamPreallocDivisor;
  SetLength(BINBuffer, 0);

  // write textual XML representation to a stream, ...
  XMLStream := TMemoryStream.Create;
  XMLStream.SetSize(PreallocSize);
  XMLDocument.SaveToStream(XMLStream);

  // ... free up memory ...
  XMLDocument.Free;

  // ... and copy to a plain buffer
  SetLength(XMLBuffer, XMLStream.Position);
  XMLStream.Position := 0;
  XMLStream.Read(XMLBuffer[0], Length(XMLBuffer));

  XMLStream.Clear;
  XMLStream.Free;

  Result := True
end;

end.
