{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/convert/ConvXML.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2017-09-26 15:58:11 +0200 (Di, 26 Sep 2017) $
  Revision: $Rev: 462 $

  Visionaire Engine converter for textual XMLs
}
unit ConvXML;

interface

uses
  Shared;

function ConvertXMLToVED(const XMLBuffer: TBuffer;
  out VEDBuffer: TBuffer): Boolean;

implementation

uses
  MemSearch,
  VEDMap,
  VEDMapEx,
  NativeXml,
  Classes,
  StrUtils,
  SysUtils;

function FixContent(const AttributeValue: WideString): WideString;
const
  Pattern: Char = 'T';
var
  i,
  Idx,
  BaseIdx,
  Len        : Longint;
  Ident      : WideString;
  IsDataField: Boolean;
begin
  Result := AttributeValue;

  BaseIdx := 1;
  repeat
    Idx := SearchMem(SearchMethod_KnuthMorrisPratt, @Pattern,
      @Result[BaseIdx], 1, (Length(Result) - BaseIdx + 1) shl 1);

    if (Idx >= 0) and (Idx and 1 = 0) then
    begin
      Idx := Idx shr 1;

      // Check whether on the left of our position is either nothing (i.e. we're
      // at the beginning of the attribute value) or there's some character
      // that is certainly not part of an identifier.
      if (BaseIdx + Idx = 1) or not (Char(Result[BaseIdx + Idx - 1]) in
        ['A'..'Z', 'a'..'z', '_', '0'..'9'])
        then
      begin
        // Search for the next position which is not a numeric character.
        Len := 0;
        while (Idx + Len < Length(Result)) and
          (Char(Result[BaseIdx + Idx + Len + 1]) in ['0'..'9']) do
          Inc(Len);

        // Check whether on the right of our possible identifier there is either
        // nothing (i.e. we're at the end of the attribute value) or there's
        // some character that is certainly not part of an identifier.
        if (Len > 1) and ((BaseIdx + Idx + Len = Length(Result)) or
          not (Char(Result[BaseIdx + Idx + Len + 1]) in
          ['A'..'Z', 'a'..'z', '_', '0'..'9']))
          then
        begin
          // We assume to have found a replaceable identifier of the type "Tnnn"
          // now. Before replacing, we'll try to heuristically detect whether or
          // not the identifier takes the role of a direct data field, which has
          // to be prepended with a "V" (as in "setValue(VValueInt, ...)").
          // We do this by analyzing the direct left context and trying to find
          // structures such as "doSomething("Object.Tnnn")" or
          // "doSomething(Object["Tnnn"])", which imply that the identifier is
          // NOT a direct data field.
          // Obviously, this approach can never be perfect -- only a full Lua
          // parser would be --, but it should do the trick in most cases.
          IsDataField := True;
          i := Idx;
          while BaseIdx + i > 1 do
          begin
            Dec(i);

            if Char(Result[BaseIdx + i]) in [' ', #009] then
              // Skip some white space -- note: we should theoretically also
              // skip <br/> and such, but that might be overkill ...
              Continue
            else if Char(Result[BaseIdx + i]) in ['.', '"'] then
            begin
              IsDataField := False;
              Break
            end
            else
              Break
          end;

          Ident := MapIdentifier(Copy(Result, BaseIdx + Idx, Len + 1),
            IsDataField);

          Result := LeftStr(Result, BaseIdx + Idx - 1) + Ident +
            RightStr(Result, Length(Result) - BaseIdx - Idx - Len);

          Inc(BaseIdx, Idx + Length(Ident))
        end
        else
          Inc(BaseIdx)
      end
      else
        Inc(BaseIdx)
    end
  until Idx < 0
end;

function FixAttributeValue(const AttributeValue: WideString): WideString;
begin
  Result := FixContent(AttributeValue)
end;

procedure RenameNode(const Node: TXmlNode);
var
  i: Longint;
begin
  // rename node
  Node.Name := MapIdentifier(Node.Name);

  // rename attribute identifiers
  for i := 0 to Node.AttributeCount - 1 do
  begin
    Node.Attributes[i].Name := MapIdentifier(Node.Attributes[i].Name);
    Node.Attributes[i].Value := FixAttributeValue(Node.Attributes[i].Value)
  end;

  // rename children recursively
  for i := 0 to Node.NodeCount - 1 do
    if Node.Nodes[i].ElementType = xeElement then
      RenameNode(Node.Nodes[i])
end;

function LoadXML(out XMLDocument: TNativeXml;
  const XMLBuffer: TBuffer): Boolean;
var
  XMLStream: TMemoryStream;
begin
  Result := False;

  try
    // copy the input buffer to a stream ...
    XMLStream := TMemoryStream.Create;
    XMLStream.Write(XMLBuffer[0], Length(XMLBuffer));
    XMLStream.Position := 0;

    // ... and build the structured document from there
    XMLDocument := TNativeXml.Create(nil);
    XMLDocument.LoadFromStream(XMLStream);

    XMLStream.Free
  except
    on LoadException: Exception do
    begin
      WriteLn(LoadException.Message);
      Exit
    end
  end;

  Result := True
end;

function SaveVED(const XMLDocument: TNativeXML;
  out VEDBuffer: TBuffer): Boolean;
var
  XMLStream: TMemoryStream;
begin
  Result := False;

  try
    // write textual XML representation to a stream ...
    XMLStream := TMemoryStream.Create;
    XMLDocument.SaveToStream(XMLStream);

    // ... and copy to a plain buffer
    SetLength(VEDBuffer, XMLStream.Position);
    XMLStream.Position := 0;
    XMLStream.Read(VEDBuffer[0], Length(VEDBuffer));

    XMLStream.Free
  except
    on SaveException: Exception do
    begin
      WriteLn(SaveException.Message);
      Exit
    end
  end;

  Result := True
end;

function ConvertXMLToVED(const XMLBuffer: TBuffer;
  out VEDBuffer: TBuffer): Boolean;
var
  XMLDocument: TNativeXml;
begin
  // load the buffered XML data into a structured document
  Result := LoadXML(XMLDocument, XMLBuffer);
  if not Result then
    Exit;

  // do the conversion (recursively)
  RenameNode(XMLDocument.Root);

  // write the converted document into the output buffer
  Result := SaveVED(XMLDocument, VEDBuffer)
end;

end.
