{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/convert/Names3.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2021-08-25 23:39:26 +0200 (Mi, 25 Aug 2021) $
  Revision: $Rev: 596 $

  Visionaire Engine 3/4 handling for internal file names
}
unit Names3;

interface

uses
  Shared,
  VISFmt;

const
  ArchiveDesignatorLength = 13;

function IsArchiveDesignator(const Candidate: AnsiString): Boolean;
procedure GenerateNames(const Buffer: TBuffer; BinaryMode: Boolean);
procedure ReadNames;
procedure GetName(const VISName: AnsiString; const DirEntry: TDirEntry;
  EntryNum: Longint; out RealName, OutName: UTF8String);

implementation

uses
  CmdLine,
  MemSearch,
  Classes,
  IniFiles,
  StrUtils,
  SysUtils,
  Windows;

const
  CrLf: AnsiString = #013#010;

var
  NameTable: THashedStringList;

function IsArchiveDesignator(const Candidate: AnsiString): Boolean;
var
  Dummy: Longint;
begin
  Result := False;

  if Length(Candidate) <> ArchiveDesignatorLength then
    Exit;

  // ensure format #a#???#?????#
  if (Candidate[1] <> '#') or
    (Candidate[3] <> '#') or
    (Candidate[7] <> '#') or
    (Candidate[13] <> '#') or
    (not (Char(Candidate[2]) in ['v', 's', 'c', 'm'])) then
    Exit;

  // ensure format #a#???#nnnnn#
  if not TryStrToInt(Copy(Candidate, 8, 5), Dummy) then
    Exit;

  // ensure for a="v" that format is #a#-01#nnnnn#
  if (Candidate[2] = 'v') and
    (Copy(Candidate, 4, 3) <> '-01') then
    Exit;

  // ensure for a either "s", "c" or "m" that format is #a#nnn#nnnnn#
  if (Candidate[2] <> 'v') and
    not TryStrToInt(Copy(Candidate, 4, 3), Dummy) then
    Exit;

  // candidate has passed all checks
  Result := True
end;

procedure FindTextNames(const Buffer: TBuffer; const NameBegPattern: AnsiString;
  const NameEndPattern: AnsiString; var NameList: TStringList);
const
  // Bounds for feasible name lengths:
  // - The minimum name length would be one character plus the appended archive
  //   designator.
  // - The maximum name length would be the longest possible name on an NT
  //   system plus the length of the appended archive designator.
  MinNameLen = 1 + ArchiveDesignatorLength;
  MaxNameLen = MAX_PATH + ArchiveDesignatorLength;
var
  NameBeg,
  NameEnd: TBuffer;
  BaseIdx,
  BegIdx,
  EndIdx : Longint;
  Name   : UTF8String;
begin
  // The XML file within the game's main VIS archive contains mapping
  // information between a file's original name and its location within the
  // game's archive collection. Since parsing the XML file would be a bit of an
  // overkill here, we're just using a heuristic pattern search, which seems to
  // be working pretty well.
  // Note that additional names can be "hidden" within game scripts. This
  // applies to textual as well as binary XML files.

  // set up search patterns
  SetLength(NameBeg, Length(NameBegPattern));
  SetLength(NameEnd, Length(NameEndPattern));

  Move(NameBegPattern[1], NameBeg[0], Length(NameBegPattern));
  Move(NameEndPattern[1], NameEnd[0], Length(NameEndPattern));

  // Search the text buffer from back to front: The terminating combination #"
  // is much more significant than the opening ".
  BaseIdx := High(Buffer);
  while BaseIdx >= 0 do
  begin
    // search for the end of a name
    EndIdx := SearchMemReverse(SearchMethod_KnuthMorrisPratt, @NameEnd[0],
      @Buffer[0], Length(NameEnd), BaseIdx + 1);

    if EndIdx < 0 then
      Break;

    BaseIdx := EndIdx - 1;

    // search for the beginning of the name
    BegIdx := SearchMemReverse(SearchMethod_KnuthMorrisPratt, @NameBeg[0],
      @Buffer[0], Length(NameBeg), EndIdx + 1);

    if BegIdx < 0 then
      Continue;

    // get our candidate
    SetLength(Name, EndIdx - BegIdx);
    Move(Buffer[BegIdx + Length(NameBeg)], Name[1], Length(Name));

    // sanitize the name (keeping potential UTF-8 characters)
    Name := AnsiReplaceStr(Name, '/', '\');

    // Do a final check: If the name ends in something resembling an archive
    // designator, we'll consider it valid.
    if (Length(Name) >= MinNameLen) and
      IsArchiveDesignator(Copy(Name, Length(Name) - ArchiveDesignatorLength + 1,
        ArchiveDesignatorLength)) then
    begin
      // store the name
      NameList.Add(Name);
      BaseIdx := BegIdx - 1
    end
  end;

  SetLength(NameBeg, 0);
  SetLength(NameEnd, 0)
end;

procedure FindBinaryNames(const Buffer: TBuffer; var NameList: TStringList);
const
  // Bounds for feasible name lengths:
  // - The minimum name length would be one character plus the appended archive
  //   designator plus the zero terminator.
  // - The maximum name length would be the longest possible name on an NT
  //   system plus the length of the appended archive designator plus the zero
  //   terminator.
  MinNameLen = 1 + ArchiveDesignatorLength + 1;
  MaxNameLen = MAX_PATH + ArchiveDesignatorLength + 1;
var
  NameEnd: TBuffer;
  Size,
  BaseIdx,
  BegIdx : Longint;
  Name   : UTF8String;
begin
  // The XML file within the game's main VIS archive contains mapping
  // information between a file's original name and its location within the
  // game's archive collection. The binary variant of this XML structure can be
  // fully decoded, but the heuristic approach it a lot faster than doing that.

  // set up search patterns
  SetLength(NameEnd, 2);

  NameEnd[0] := Ord('#');
  NameEnd[1] := $00;

  // Search the binary buffer from front to back: We try to interpret any double
  // word as a string length. If the proposed string following this double word
  // fulfills two criteria, we assume it to be a valid name:
  // - The data has to end with a hash sign and a zero terminator.
  // - The data in between the proposed length and the end does not contain a
  //   zero character.
  // (Theoretically we could check for non-prinatble characters within the data.
  // However, since the names may contain non-trivial UTF-8 characters, this
  // seems a bit too complicated. We'll stay with this approach until it
  // fails, in which case checking the candidate after the UTF8-to-ANSI-
  // conversion might be helpful ...)
  BaseIdx := 0;
  while BaseIdx <= High(Buffer) - SizeOf(Size) + 1 do
  begin
    // get the proposed size
    Move(Buffer[BaseIdx], Size, SizeOf(Size));
    BegIdx := BaseIdx + SizeOf(Size);

    // Check for feasibility: If the size is too short or too long, we'll ignore
    // this candidate.
    if (Size < MinNameLen) or (Size > MaxNameLen) or
      (BegIdx + Size >= High(Buffer)) then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // First validity check: Does the name candidate end with a hash sign and a
    // zero terminator?
    if not CompareMem(@Buffer[BegIdx + Size - Length(NameEnd)], @NameEnd[0],
      Length(NameEnd)) then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // get our candidate
    SetLength(Name, Size - 1);
    Move(Buffer[BegIdx], Name[1], Length(Name));

    // Pre-final check: If the candidate name does not contain zero characters,
    // we'll consider it to be valid enough to be sanitized.
    if Pos(#000, Name) > 0 then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // sanititze the name (keeping potential UTF-8 characters)
    Name := AnsiReplaceStr(Name, '/', '\');

    // Do a final check: If the name ends in something resembling an archive
    // designator, we'll consider it valid.
    if (Length(Name) < MinNameLen) or not
      IsArchiveDesignator(Copy(Name, Length(Name) - ArchiveDesignatorLength + 1,
        ArchiveDesignatorLength)) then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // store the name
    NameList.Add(Name);

    BaseIdx := BegIdx + Size
  end;

  SetLength(NameEnd, 0)
end;

procedure GenerateNames(const Buffer: TBuffer; BinaryMode: Boolean);
var
  NameFile: file;
  i       : Longint;
  NameList: TStringList;
begin
  // open target name table file
  if not OpenOutFile(NameFile, ParamRec.NameSpec) then
  begin
    WriteLn('    error opening name file');
    Exit
  end;

  // get file names
  NameList := TStringList.Create;
  NameList.CaseSensitive := True;
  NameList.Sorted := False;

  // When searching for names, we usually search for archive designators
  // appended to the names. We have to differentiate two cases:
  // - In "classical" textual XML files, names are enclosed in regular double
  //   quotes when appearing as an attribute value. They can also appear within
  //   attribute values (e.g. scripts). In that case, the quotes are encoded.
  // - In binary XML files, names that are attribute values are stored on a
  //   binary level (using a prepended length value). Names that appear within
  //   an attribute value can now be found within unencoded quotes.
  //
  // This especially means that we can search for names contained in unencoded
  // quotes in any case.
  FindTextNames(Buffer, '"', '#"', NameList);

  if BinaryMode then
    FindBinaryNames(Buffer, NameList)
  else
    FindTextNames(Buffer, '&quot;', '#&quot;', NameList);

  // sort the list to easily find duplicates
  NameList.Sort;

  // output all unique names
  for i := 0 to NameList.Count - 1 do
    if (i = 0) or (NameList[i] <> NameList[i - 1]) then
    begin
      {$I-}
      BlockWrite(NameFile, NameList[i][1], Length(NameList[i]));
      BlockWrite(NameFile, CrLf[1], Length(CrLf));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    error writing to name file');
        Break
      end
    end;

  NameList.Free;
  CloseFile(NameFile)
end;

procedure ReadNames;
var
  Count,
  p    : Longint;
  Names: TextFile;
  Line,
  Ident: UTF8String;
begin
  AssignFile(Names, ParamRec.NameSpec);
  {$I-}
  Reset(Names);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('warning: cannot read name file');
    Exit
  end;

  // count lines
  Count := 0;
  while not Eof(Names) do
  begin
    Inc(Count);
    ReadLn(Names)
  end;

  // prepare table and read names
  NameTable.Capacity := Count;

  Reset(Names);
  while not Eof(Names) do
  begin
    ReadLn(Names, Line);

    // search for the fourth "#" from the end of the name -- this will serve as
    // our identifier
    Ident := AnsiReverseString(Line);
    p := PosEx('#', Ident, 1);
    if p > -1 then
      p := PosEx('#', Ident, p + 1);
    if p > -1 then
      p := PosEx('#', Ident, p + 1);
    if p > -1 then
      p := PosEx('#', Ident, p + 1);

    if p > - 1 then
    begin
      Line := RightStr(Line, p) + '=' + LeftStr(Line, Length(Line) - p);
      NameTable.Add(Line)
    end
  end;

  CloseFile(Names)
end;

procedure GetName(const VISName: AnsiString; const DirEntry: TDirEntry;
  EntryNum: Longint; out RealName, OutName: UTF8String);
var
  i      : Longint;
  FileExt,
  Group,
  Count,
  Mask   : AnsiString;
begin
  RealName := '';
  OutName := '';

  // if a name table has been loaded, try to get the file's real name
  if ParamRec.UseNames then
  begin
    // Each entry in the name table is marked with an identifier that
    // corresponds to the archive file extension:
    // .vs035, file number 12 --> #s#035#00012#
    FileExt := AnsiLowerCase(ExtractFileExt(VISName));

    if FileExt = '.vis' then
    begin
      // the main VIS archive has the special designation "#v#-01#"
      Group := 'v';
      Count := '-01';
    end
    else if FileExt = '.vc' then
    begin
      // special case for only a single vc file: "#c#-01#"
      Group := 'c';
      Count := '-01';
    end
    else if FileExt = '.vi' then
    begin
      // special case for only a single vi file: "#i#-01#"
      Group := 'i';
      Count := '-01';
    end
    else if FileExt = '.vo' then
    begin
      // special case for only a single vo file: "#o#-01#"
      Group := 'o';
      Count := '-01';
    end
    else if FileExt = '.vs' then
    begin
      // special case for only a single vs file: "#s#-01#"
      Group := 's';
      Count := '-01';
    end
    else if Copy(FileExt, 1, 2) = '.v' then
    begin
      // scene or character files follow the standard designation
      Group := Copy(FileExt, 3, 1);
      Count := Copy(FileExt, 4, Length(FileExt) - 3);

      // special case for video files
      if Group = 'v' then
        Group := 'm'
    end
    else
      Group := '';

    if Group <> '' then
    begin
      Mask := Format('#%s#%s#%.5d#', [Group, Count, EntryNum]);

      i := NameTable.IndexOfName(Mask);
      if i >= 0 then
      begin
        RealName := NameTable.Values[Mask];
        OutName := RealName;

        if ParamRec.Unique then
        begin
          FileExt := ExtractFileExt(OutName);
          OutName := Copy(OutName, 1, Length(OutName) - Length(FileExt)) +
            Mask + FileExt
        end
        else if ParamRec.UseDefaultName then
          OutName := ''
      end
    end
  end;

  // if no friendly name could be found, generate a default
  if OutName = '' then
    with DirEntry do
    begin
      OutName := Format('%s_%.5d_%s',
        [VISName, EntryNum, AnsiLowerCase(IntToHex(Offset, 8))]);

      if (Flags and Flag_VEBMainXML) <> 0 then
        OutName := OutName + '.veb'
      else if (Flags and Flag_VEDMainXML) <> 0 then
        OutName := OutName + '.ved'
      else if (Flags and Flag_BinaryMainXML) <> 0 then
        OutName := OutName + '.vbin'
      else if (Flags and Flag_TextMainXML) <> 0 then
        OutName := OutName + '.xml'
      else if (Flags and Flag_CryptImage) <> 0 then
        // we're guessing PNG for now, but it might be a WebP instead ...
        OutName := OutName + '.png'
      else
        OutName := OutName + '.dat'
    end
end;

initialization
  NameTable := THashedStringList.Create

finalization
  if Assigned(NameTable) then
    NameTable.Free

end.
