{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/convert/Names5.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2023-05-30 01:13:35 +0200 (Di, 30 Mai 2023) $
  Revision: $Rev: 659 $

  Visionaire Engine 5 handling for internal file names
}
unit Names5;

interface

uses
  Shared,
  VISFmt;

const
  ArchiveDesignatorLength = 12;

function IsArchiveDesignator(const Candidate: AnsiString): Boolean;
procedure GenerateNames(const Buffer: TBuffer; BinaryMode: Boolean);
procedure ReadNames;
procedure GetName(const VISName: AnsiString; const DirEntry: TDirEntry;
  EntryNum: Longint; out RealName, OutName: UTF8String);

implementation

uses
  ConvVBIN,
  CmdLine,
  MemSearch,
  Classes,
  IniFiles,
  NativeXml,
  StrUtils,
  SysUtils,
  Windows;

const
  CrLf: AnsiString = #013#010;

var
  NameTable   : THashedStringList;
  ArchiveTable: THashedStringList;

function GenerateArchiveNames(const XMLDocument: TNativeXml;
  var NameFile: file): Boolean;
var
  i   : Longint;
  Node: TXmlNode;
  Name: UTF8String;
begin
  Result := False;

  // follow path to the game container names
  Node := XMLDocument.Root;

  if Assigned(Node) then
    Node := Node.NodeByName('T116');

  if Assigned(Node) then
    Node := Node.NodeByName('T803');

  if not Assigned(Node) then
  begin
    WriteLn('    error determining game container structure');
    Exit
  end;

  // store the number of container names
  Name := IntToStr(Node.NodeCount);
  {$I-}
  BlockWrite(NameFile, Name[1], Length(Name));
  BlockWrite(NameFile, CrLf[1], Length(CrLf));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('    error writing to name file');
    Exit
  end;

  // store container names
  for i := 0 to Node.NodeCount - 1 do
  begin
    if not Assigned(Node.Nodes[i].NodeByName('T25')) then
    begin
      WriteLn('    error determining game container names');
      Exit
    end;

    Name := Node.Nodes[i].NodeByName('T25').Value;

    // The game archives might have been compiled into subdirectories. We remove
    // the directory names here, since they will not be known when processing a
    // single archive.
    Name := ExtractFileName(Name);

    {$I-}
    BlockWrite(NameFile, Name[1], Length(Name));
    BlockWrite(NameFile, CrLf[1], Length(CrLf));
    {$I+}
    if IOResult <> 0 then
    begin
      WriteLn('    error writing to name file');
      Exit
    end
  end;

  Result := True
end;

function LoadXML(out XMLDocument: TNativeXml;
  const XMLBuffer: TBuffer): Boolean;
var
  XMLStream: TMemoryStream;
begin
  Result := False;

  try
    // copy the input buffer to a stream ...
    XMLStream := TMemoryStream.Create;
    XMLStream.Write(XMLBuffer[0], Length(XMLBuffer));
    XMLStream.Position := 0;

    // ... and build the structured document from there
    XMLDocument := TNativeXml.Create(nil);
    XMLDocument.LoadFromStream(XMLStream);

    XMLStream.Free
  except
    on LoadException: Exception do
    begin
      WriteLn(LoadException.Message);
      Exit
    end
  end;

  Result := True
end;

function IsArchiveDesignator(const Candidate: AnsiString): Boolean;
var
  Dummy: Longint;
begin
  Result := False;

  if Length(Candidate) <> ArchiveDesignatorLength then
    Exit;

  // ensure format #???#??????#
  if (Candidate[1] <> '#') or
    (Candidate[5] <> '#') or
    (Candidate[12] <> '#') then
    Exit;

  // ensure format #nnn#nnnnnn# or #nnn#mnnnnn#
  if not TryStrToInt(Copy(Candidate, 2, 3), Dummy) then
    Exit;
  if Candidate[6] = 'm' then
  begin
    if not TryStrToInt(Copy(Candidate, 7, 5), Dummy) then
      Exit
  end
  else
    if not TryStrToInt(Copy(Candidate, 6, 6), Dummy) then
      Exit;

  // candidate has passed all checks
  Result := True
end;

procedure FindTextNames(const Buffer: TBuffer; const NameBegPattern: AnsiString;
  const NameEndPattern: AnsiString; var NameList: TStringList);
const
  // Bounds for feasible name lengths:
  // - The minimum name length would be one character plus the appended archive
  //   designator.
  // - The maximum name length would be the longest possible name on an NT
  //   system plus the length of the appended archive designator.
  MinNameLen = 1 + ArchiveDesignatorLength;
  MaxNameLen = MAX_PATH + ArchiveDesignatorLength;
var
  NameBeg,
  NameEnd: TBuffer;
  BaseIdx,
  BegIdx,
  EndIdx : Longint;
  Name   : UTF8String;
begin
  // The XML file within the game's main VIS archive contains mapping
  // information between a file's original name and its location within the
  // game's archive collection. We decode the textual variant of the XML
  // structure to get the game container names, but use a faster heuristic
  // approach to construct a list of the actual file names.
  // Note that additional names can be "hidden" within game scripts. This
  // applies to textual as well as binary XML files.

  // set up search patterns
  SetLength(NameBeg, Length(NameBegPattern));
  SetLength(NameEnd, Length(NameEndPattern));

  Move(NameBegPattern[1], NameBeg[0], Length(NameBegPattern));
  Move(NameEndPattern[1], NameEnd[0], Length(NameEndPattern));

  // Search the text buffer from back to front: The terminating combination #"
  // is much more significant than the opening ".
  BaseIdx := High(Buffer);
  while BaseIdx >= 0 do
  begin
    // search for the end of a name
    EndIdx := SearchMemReverse(SearchMethod_KnuthMorrisPratt, @NameEnd[0],
      @Buffer[0], Length(NameEnd), BaseIdx + 1);

    if EndIdx < 0 then
      Break;

    BaseIdx := EndIdx - 1;

    // search for the beginning of the name
    BegIdx := SearchMemReverse(SearchMethod_KnuthMorrisPratt, @NameBeg[0],
      @Buffer[0], Length(NameBeg), EndIdx + 1);

    if BegIdx < 0 then
      Continue;

    // get our candidate
    SetLength(Name, EndIdx - BegIdx);
    Move(Buffer[BegIdx + Length(NameBeg)], Name[1], Length(Name));

    // sanitize the name (keeping potential UTF-8 characters)
    Name := AnsiReplaceStr(Name, '/', '\');

    // Do a final check: If the name ends in something resembling an archive
    // designator, we'll consider it valid.
    if (Length(Name) >= MinNameLen) and
      IsArchiveDesignator(Copy(Name, Length(Name) - ArchiveDesignatorLength + 1,
        ArchiveDesignatorLength)) then
    begin
      // store the name
      NameList.Add(Name);
      BaseIdx := BegIdx - 1
    end
  end;

  SetLength(NameBeg, 0);
  SetLength(NameEnd, 0)
end;

procedure FindBinaryNames(const Buffer: TBuffer; var NameList: TStringList);
const
  // Bounds for feasible name lengths:
  // - The minimum name length would be one character plus the appended archive
  //   designator plus the zero terminator.
  // - The maximum name length would be the longest possible name on an NT
  //   system plus the length of the appended archive designator plus the zero
  //   terminator.
  MinNameLen = 1 + ArchiveDesignatorLength + 1;
  MaxNameLen = MAX_PATH + ArchiveDesignatorLength + 1;
var
  NameEnd: TBuffer;
  Size,
  BaseIdx,
  BegIdx : Longint;
  Name   : UTF8String;
begin
  // The XML file within the game's main VIS archive contains mapping
  // information between a file's original name and its location within the
  // game's archive collection. We decode the binary variant of the XML
  // structure to get the game container names, but use a faster heuristic
  // approach to construct a list of the actual file names.

  // set up search patterns
  SetLength(NameEnd, 2);

  NameEnd[0] := Ord('#');
  NameEnd[1] := $00;

  // Search the binary buffer from front to back: We try to interpret any double
  // word as a string length. If the proposed string following this double word
  // fulfills two criteria, we assume it to be a valid name:
  // - The data has to end with a hash sign and a zero terminator.
  // - The data in between the proposed length and the end does not contain a
  //   zero character.
  // (Theoretically we could check for non-prinatble characters within the data.
  // However, since the names may contain non-trivial UTF-8 characters, this
  // seems a bit too complicated. We'll stay with this approach until it
  // fails, in which case checking the candidate after the UTF8-to-ANSI-
  // conversion might be helpful ...)
  BaseIdx := 0;
  while BaseIdx <= High(Buffer) - SizeOf(Size) + 1 do
  begin
    // get the proposed size
    Move(Buffer[BaseIdx], Size, SizeOf(Size));
    BegIdx := BaseIdx + SizeOf(Size);

    // Check for feasibility: If the size is too short or too long, we'll ignore
    // this candidate.
    if (Size < MinNameLen) or (Size > MaxNameLen) or
      (BegIdx + Size >= High(Buffer)) then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // First validity check: Does the name candidate end with a hash sign and a
    // zero terminator?
    if not CompareMem(@Buffer[BegIdx + Size - Length(NameEnd)], @NameEnd[0],
      Length(NameEnd)) then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // get our candidate
    SetLength(Name, Size - 1);
    Move(Buffer[BegIdx], Name[1], Length(Name));

    // Pre-final check: If the candidate name does not contain zero characters,
    // we'll consider it to be valid enough to be sanitized.
    if Pos(#000, Name) > 0 then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // sanititze the name (keeping potential UTF-8 characters)
    Name := AnsiReplaceStr(Name, '/', '\');

    // Do a final check: If the name ends in something resembling an archive
    // designator, we'll consider it valid.
    if (Length(Name) < MinNameLen) or not
      IsArchiveDesignator(Copy(Name, Length(Name) - ArchiveDesignatorLength + 1,
        ArchiveDesignatorLength)) then
    begin
      Inc(BaseIdx);
      Continue
    end;

    // store the name
    NameList.Add(Name);

    BaseIdx := BegIdx + Size
  end;

  SetLength(NameEnd, 0)
end;

procedure GenerateNames(const Buffer: TBuffer; BinaryMode: Boolean);
var
  XMLDocument: TNativeXml;
  NameFile   : file;
  i          : Longint;
  NameList   : TStringList;
begin
  // open target name table file
  if not OpenOutFile(NameFile, ParamRec.NameSpec) then
  begin
    WriteLn('    error opening name file');
    Exit
  end;

  // convert buffer data to a regular XML structure
  if BinaryMode then
  begin
    if not ConvertVBINToStructuredXML(Buffer, XMLDocument, False) then
    begin
      WriteLn('    error converting binary XML');

      CloseFile(NameFile);
      Exit
    end
  end
  else
  begin
    if not LoadXML(XMLDocument, Buffer) then
    begin
      WriteLn('    error loading textual XML');

      CloseFile(NameFile);
      Exit
    end
  end;

  // store the container names first
  if not GenerateArchiveNames(XMLDocument, NameFile) then
  begin
    CloseFile(NameFile);
    Exit
  end;

  // get file names
  NameList := TStringList.Create;
  NameList.CaseSensitive := True;
  NameList.Sorted := False;

  // When searching for names, we usually search for archive designators
  // appended to the names. We have to differentiate two cases:
  // - In "classical" textual XML files, names are enclosed in regular double
  //   quotes when appearing as an attribute value. They can also appear within
  //   attribute values (e.g. scripts). In that case, the quotes are encoded.
  // - In binary XML files, names that are attribute values are stored on a
  //   binary level (using a prepended length value). Names that appear within
  //   an attribute value can now be found within unencoded quotes.
  //
  // This especially means that we can search for names contained in unencoded
  // quotes in any case.
  FindTextNames(Buffer, '"', '#"', NameList);

  if BinaryMode then
  begin
    FindBinaryNames(Buffer, NameList);
    FindTextNames(Buffer, '''', '#''', NameList)
  end
  else
  begin
    FindTextNames(Buffer, '&quot;', '#&quot;', NameList);
    FindTextNames(Buffer, '&apos;', '#&apos;', NameList)
  end;

  // sort the list to easily find duplicates
  NameList.Sort;

  // output all unique names
  for i := 0 to NameList.Count - 1 do
    if (i = 0) or (NameList[i] <> NameList[i - 1]) then
    begin
      {$I-}
      BlockWrite(NameFile, NameList[i][1], Length(NameList[i]));
      BlockWrite(NameFile, CrLf[1], Length(CrLf));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    error writing to name file');
        Break
      end
    end;

  NameList.Free;
  CloseFile(NameFile)
end;

procedure ReadNames;
var
  Count,
  ArcCount,
  p       : Longint;
  Names   : TextFile;
  Line,
  Ident   : UTF8String;
begin
  AssignFile(Names, ParamRec.NameSpec);
  {$I-}
  Reset(Names);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('warning: cannot read name file');
    Exit
  end;

  // count lines
  Count := 0;
  while not Eof(Names) do
  begin
    Inc(Count);
    ReadLn(Names)
  end;

  // start over from the beginning and read the container count
  Reset(Names);
  ReadLn(Names, Line);
  ArcCount := StrToInt(Line);

  // prepare archive table and read container names
  ArchiveTable.Capacity := ArcCount;
  p := 0;
  while not Eof(Names) and (p < ArcCount) do
  begin
    ReadLn(Names, Line);

    // we'll use a normalized container name as an identifier
    Line := AnsiLowerCase(Line) + '=' + Line;
    ArchiveTable.Add(Line);

    Inc(p)
  end;

  // prepare table and read names
  NameTable.Capacity := Count - ArcCount - 1;

  while not Eof(Names) do
  begin
    ReadLn(Names, Line);

    // search for the third "#" from the end of the name -- this will serve as
    // our identifier
    Ident := AnsiReverseString(Line);
    p := PosEx('#', Ident, 1);
    if p > -1 then
      p := PosEx('#', Ident, p + 1);
    if p > -1 then
      p := PosEx('#', Ident, p + 1);

    if p > - 1 then
    begin
      Line := RightStr(Line, p) + '=' + LeftStr(Line, Length(Line) - p);
      NameTable.Add(Line)
    end
  end;

  CloseFile(Names)
end;

procedure GetName(const VISName: AnsiString; const DirEntry: TDirEntry;
  EntryNum: Longint; out RealName, OutName: UTF8String);
var
  i,
  ArcIdx : Longint;
  FileExt,
  Mask   : AnsiString;
begin
  RealName := '';
  OutName := '';

  // if a name table has been loaded, try to get the file's real name
  if ParamRec.UseNames then
  begin
    // Each entry in the name table is marked with an identifier that
    // corresponds to the archive number (as definied in the main XML file)
    // supplemented with the number of the file within the archive.
    // Example: archive 35, file number 12 --> #035#000012#
    ArcIdx := ArchiveTable.IndexOfName(AnsiLowerCase(VISName)) + 1;

    if ArcIdx >= 1 then
    begin
      Mask := Format('#%.3d#%.6d#', [ArcIdx, EntryNum]);

      i := NameTable.IndexOfName(Mask);
      if i >= 0 then
      begin
        RealName := NameTable.Values[Mask];
        OutName := RealName;

        if ParamRec.Unique then
        begin
          FileExt := ExtractFileExt(OutName);
          OutName := Copy(OutName, 1, Length(OutName) - Length(FileExt)) +
            Mask + FileExt
        end
        else if ParamRec.UseDefaultName then
          OutName := ''
      end
    end
  end;

  // if no friendly name could be found, generate a default
  if OutName = '' then
    with DirEntry do
    begin
      OutName := Format('%s_%.5d_%s',
        [VISName, EntryNum, AnsiLowerCase(IntToHex(Offset, 8))]);

      if (Flags and Flag_VEBMainXML) <> 0 then
        OutName := OutName + '.veb'
      else if (Flags and Flag_VEDMainXML) <> 0 then
        OutName := OutName + '.ved'
      else if (Flags and Flag_BinaryMainXML) <> 0 then
        OutName := OutName + '.vbin'
      else if (Flags and Flag_TextMainXML) <> 0 then
        OutName := OutName + '.xml'
      else if (Flags and Flag_CryptImage) <> 0 then
        // we're guessing PNG for now, but it might be a WebP instead ...
        OutName := OutName + '.png'
      else
        OutName := OutName + '.dat'
    end
end;

initialization
  NameTable := THashedStringList.Create;
  ArchiveTable := THashedStringList.Create

finalization
  if Assigned(NameTable) then
    NameTable.Free;
  if Assigned(ArchiveTable) then
    ArchiveTable.Free

end.
